# -----------------------------------------------------------------------------------------
#
#   IGES export for SketchUp 8
#
#   Copyright 2012, Rainer Ochs
#
#	Data: 2012-01-23
#	Author: Rainer Ochs
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#   Permission to use this software for any purpose and without fee is hereby granted
#   Distribution of this software for commercial purpose only with the the expressed and written consent of the author
#   If you ditribute this software you have to include this copyright notice in all copies.
#	contact: rainer.ochs@gmx.de   If you want to make a donation, use paypal with this mail address
#
#   Description
#   The program is a Ruby script for SketchUp and is installed by copying into the plugins directory
#	It will create a function "IGES Export" in the Tools Menu
#	The software will export the Sketchup geometry into a IGES file that should be readable by CAD programs
#	The geometry is expressed as a manifold boundary-representation (BREP) in IGES
# 	Some checking is done if the geometry represents valid solid geometry
#   Every facet of the sketchup model is exported as one face to the BREP. This also applies to the curved geometry.
#	A cylinder is not one surface shell but a multitude of curved segments forming the cylinder shell.
#
#	Version:
#   V 0.1	First implementation. All geometry is exported as edgy facets.
#	V 0.2	speedup, remove construction lines. Remove of non-solid elements before export.
#	V 0.3	Export of cylinders and cones.
#			This is still a facetted model, but the facets nor are no longer flat but take the correct shape.
#			other shapes are still flat.
#   V 0.4	first release
#   V 0.5	undo corrected
#   V 0.6	deletes temp file. Removes inner faces instead of deleting the whole object.
#			Clean exit if no filename or nothing to export
#   V 0.7	Improved preprocessing - removes edges that connect to inner faces.
#--------------------------------------------------------------------------------------------

require 'sketchup.rb'

module IGESexport

  @filename = "";			# export filename
  @fileParams = "";			# tmporary filename for params
  @dataline = ""			# line for sequential filed
  @lastType = 0;			# current parameter type
  @objects = Array.new()	# all entities in a nested array one subarry per connected object

  # constant definitions for IGES STATUS field
  BLANK_VISIBLE = 00
  BLANK_BLANKED = 01
  SUB_INDEPENDANT = 00
  SUB_PHYS_DEPEND = 01
  SUB_LOG_DEPEND = 02
  SUB_BOTH = 03
  USE_GEOMETRY = 00
  USE_ANNOTATION = 01
  USE_DEFINITION = 02
  USE_OTHER = 03
  USE_LOGICAL = 04
  USE_PARAMETRIC = 05
  USE_CONSTRUCT = 06
  HIRARCHY_TOPDOWN = 00
  HIRARCHY_DEFER = 01


  def IGESexport.export2IGES()

    model = Sketchup.active_model

    # get selected/all entities of the active model
    if model.selection.empty?
      entities = model.entities
    else
      entities = model.selection
    end

    if entities.length == 0
      UI.messagebox("Nothing to do!")
      return nil
    end

    # define the filename
    modelName = "Untitled" if (modelName = File.basename(model.path, ".skp")) == ""

    model.start_operation("IGES export preparation", true)

	Sketchup.status_text="IGES Export ..."

	removeAllAttributes(Sketchup.active_model.entities)


	while true
	   @clean = true
	   eraseInnerFaces(Sketchup.active_model.entities)
	   break if @clean
	end    # the each-loop has to be redone after an erase

	@erased = false
	while true
	   @clean = true
	   checkSolid(Sketchup.active_model.entities)
	   break if @clean
	end    # the each-loop has to be redone after an erase

	if @erased
		UI.messagebox("Model contains non-solid elements, these will not be exported.")
	end
    @objects.clear;
    copyEntities(Sketchup.active_model.entities)
	if @objects.size == 0
		UI.messagebox("Model does not contain solid elements, nothing will not be exported.")
		model.commit_operation
		Sketchup.undo
		return

	end


    @filename = UI.savepanel( "Save IGES file ...", File.dirname(model.path) , "#{modelName}.igs" )
	if not @filename
		model.commit_operation
		Sketchup.undo
		return
	end
	@fileParams = @filename + '.tmp'

    doExport2IGES()

    model.commit_operation
    Sketchup.undo

    model.selection.clear()

    Sketchup.set_status_text("done")

  end


  def IGESexport.copyEntities(entities)
	# set all groups to zero transformation by applying the transformation to the entities
	# make all component instances unique and also change the transformation
	entities.each { |e|
	    a = e.get_attribute("IGES","Object")
		if (not e.deleted?) && (not e.hidden?) & (a == nil)
			if e.is_a? Sketchup::ComponentInstance
				e.make_unique
				tr = e.transformation
				e.definition.entities.transform_entities(tr, e.definition.entities.to_a)
				e.transformation=Geom::Transformation.new()				# set zero transformation
				copyEntities(e.definition.entities)						# recursively do the same for all group entities
			elsif e.is_a? Sketchup::Group
				e.make_unique
				tr = e.transformation
				e.entities.transform_entities(tr, e.entities.to_a)
				e.transformation=Geom::Transformation.new()				# set zero transformation
				copyEntities(e.entities)									# recursively do the same for all group entities
			else
				obj = Array.new
				e.all_connected.each { |e|
                    if (e.typename == "Edge")  or (e.typename == "Face")
						obj << e
					    e.set_attribute("IGES","Object","valid")
					end
                }
				@objects << obj if obj.length > 0
            end
		end
	}
end

  def IGESexport.eraseInnerFaces(entities)
	# check if the model contains only solid geometry. Lines that are not connectd to two faces are deleted
	# along with all connected entities.
	# this has to be done recursively to check all groups and components
	entities.each { |f|
		if (not f.deleted?) && (not f.hidden?)
			if (f.typename == "Face")
			    inner = true;
			    f.edges.each { |e|
					if e.faces.size < 3			# less then three faces, this is not inner face.
						inner = false
					end
				}
				if inner == true
					entities.erase_entities f
					@clean = false
				end
			elsif f.is_a? Sketchup::ComponentInstance
				eraseInnerFaces(f.definition.entities)						# recursively do the same for all group entities
			elsif f.is_a? Sketchup::Group
				eraseInnerFaces(f.entities)									# recursively do the same for all group entities
            end
		end
	}
end

  def IGESexport.checkSolid(entities)
	# check if the model contains only solid geometry. Lines that are not connectd to two faces are deleted
	# along with all connected entities.
	# this has to be done recursively to check all groups and components
	entities.each { |e|
		if (not e.deleted?) && (not e.hidden?)
			if ((e.typename == "ConstructionLine") or
				(e.typename == "ConstructionPoint") or
				(e.typename == "DimensionLinear") or
				(e.typename == "SectionPlane") or
				(e.typename == "DimensionRadial" ))
					entities.erase_entities e
				@clean = false

			elsif (e.typename == "Edge")
				if e.faces.size != 2				# not two faces, this may be some lines or faces connected with an valid object.
													# or represent inner geometry
					entities.erase_entities e
					@erased = true
					@clean = false;
				end
			elsif e.is_a? Sketchup::ComponentInstance
				checkSolid(e.definition.entities)						# recursively do the same for all group entities
			elsif e.is_a? Sketchup::Group
				checkSolid(e.entities)									# recursively do the same for all group entities
            end
		end
	}
end

# checks the entities for cylindrical and conical
# the geometry is extracted and added as parameters to faces and edges.
def IGESexport.geometryTest(entities)
	entities.each {|f|								# for all model entities
		if (f.typename == "Face") && (not f.deleted?) && (not f.hidden?)	# only for valid faces
			f.checkGeometryCone()												# do this face
		end
	}
end

# create a IGES directory entry
# type = entity type
# param is the starting line number bin the parameter section
# count is the line count this parameter occupys
# fields not given are defaulted
  def IGESexport.genDirectoryEntry(type,param,structure,font,status,blank,sub,use,hir,count,form,trans=0)
	printf(@fileDE, "%8d%8d%8d%8d%8d%8d%8d%8d%02d%02d%02d%02dD%7d\n",type,param,structure,font,0,0,trans,0,blank,sub,use,hir,@seqDE)									# number of dimensions
	@seqDE += 1;
	printf(@fileDE, "%8d%8d%8d%8d%8d%8d%8d%8d%8dD%7d\n",type,0,0,count,form,0,0,0,0,@seqDE)
	@seqDE += 1;
  end


# create a IGES parameter entry in tempfile
# this call will be called until all data is saved. Lines will be stuffed automatically
# type is the entity type this params are for. If zero indicates that no more data will follow
# num is the number of data blocks that will be saved
# data is the data as a string.
  def IGESexport.addParamData(type,data)
	if (type == 0)
        delimiter = ";"
	else
		delimiter = ","
    end
	if (type != @lastType) and (type > 0)  # a new Entry is starting
	    @dataline = type.to_s + delimiter +data+delimiter
		@lastType = type;
	else
		if @dataline.size + data.size >= 63		# does not fit in one line
			printf(@fileP, "%-64s%8dP%7d\n",@dataline,@seqDE,@seqP)
			@dataline = data +delimiter
			@seqP += 1;
		else
			@dataline += data +delimiter
		end
	end
	if (type == 0)
		printf(@fileP, "%-64s%8dP%7d\n",@dataline,@seqDE,@seqP)
		@dataline = data +delimiter
		@seqP += 1;
		@lastType = 0;
	end
end

 def IGESexport.doExport2IGES()
  	@fileDE = File.new( @filename, "w" )
	printf(@fileDE,"                                                                        S0000001\n")	# start section
	printf(@fileDE,",,22HSketchUp IGES exporter,13HFilename.iges,4HV0.1,4HV0.1,32,          G0000001\n")	# global section
	printf(@fileDE,",38,6,99,15,1.,2,2HMM,1,0.01,15H20120122.221849,1E-007,10.,6Hxxxxxx,,   G0000002\n")
	printf(@fileDE,"11,0,15H20120122.221849,;                                               G0000003\n")

	@fileP = File.new( @fileParams, "w" )

    @objects.each{|obj|
		geometryTest(obj)
		entities110_edges(obj)				# export all edges to a entity 110 entry
		entities100_arcs(obj)				# export all arcs to a entity 100 entry
		entity502_Vertices(obj)				# export all vertices in a vertices list entity 502
		entities504_edgeList(obj);
		entities508_loops(obj);
		entities190_surfaces(obj)
		entities510_faces(obj)
		entities514_FaceList(obj)
		genManifold_186()
		}

    # add the params temp file to the main file
	@fileP.close
		File.open(@fileParams) { |f|
		  f.each_line { |line|									# process one line after the other
             printf(@fileDE,line)
			 }
	}
	printf(@fileDE, "S      1G      3D%7dP%7d                                        T0000001\n",@seqDE-1,@seqP-1)
	@fileDE.close;
    File.delete(@fileParams)
  end

  def IGESexport.removeAllAttributes(e)
	e.each {|f|						# for all model entities
	    f.delete_attribute("IGES")
		if (f.typename == "Edge") && (not f.deleted?) && (not f.hidden?)   #
			v = f.vertices[0];
			v.delete_attribute("IGES")
			v = f.vertices[1];
			v.delete_attribute("IGES")
		end
		if (f.typename == "Face") && (not f.deleted?) && (not f.hidden?)   #
			f.loops.each{|l|
			  l.delete_attribute("IGES")
			}
		end
		if (f.typename == "Group") && (not f.deleted?) && (not f.hidden?)
			removeAllAttributes(f.entities)
		end
		if (f.typename == "ComponentInstance") && (not f.deleted?) && (not f.hidden?)
			removeAllAttributes(f.definition.entities)
		end
	}
	@ParamE = 1;
	@seqDE = 1;
	@seqP = 1;
  end


  def IGESexport.entities110_edges(e)
	edgeListIndex = 1;
	# first count the vertices and assign indices
	e.each {|f|						# for all model entities
		if (f.typename == "Edge") && (not f.deleted?) && (not f.hidden?)   # valid face ?
			if f.get_attribute("IGES","type") == nil   		# edge is straight
				f.set_attribute("IGES","DE",@seqDE.to_s)  		# Directory Entry of Vertices 502
				edgeListIndex += 1;
				# Create param entry
				startParamLine = @seqP
				p = f.vertices[0].position
				addParamData(110,p[0].to_mm.to_f.to_s)
				addParamData(110,p[1].to_mm.to_f.to_s)
				addParamData(110,p[2].to_mm.to_f.to_s)
				p = f.vertices[1].position
				addParamData(110,p[0].to_mm.to_f.to_s)
				addParamData(110,p[1].to_mm.to_f.to_s)
				addParamData(0,p[2].to_mm.to_f.to_s)
				# now generate the matching directory entry
				genDirectoryEntry(110,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);
			end
		end
	}
	@numEdges = edgeListIndex-1
  end

  def IGESexport.entities100_arcs(e)
	edgeListIndex = 1;
	# first count the vertices and assign indices
	e.each {|f|						# for all model entities
		if (f.typename == "Edge") && (not f.deleted?) && (not f.hidden?)   # valid face ?
			if f.get_attribute("IGES","type") == "arc"   		# edge is an arc
				edgeListIndex += 1;
                m1 = Geom::Point3d.new(f.get_attribute("IGES","mx").to_f,f.get_attribute("IGES","my").to_f,f.get_attribute("IGES","mz").to_f)
				# create transformation entry
				p1 = f.vertices[0].position
				p2 = f.vertices[1].position
				v1 = m1.vector_to(p1)
				v2 = m1.vector_to(p2)
				v = v1.cross(v2)				# a vector perpendicular to the plane of the circle
				tr = Geom::Transformation.new(Geom::Point3d.new(0,0,0), v).inverse

				startParamLine = @seqP
				for i in 0..10 do
					addParamData(124,tr.to_a[i].to_f.to_s)
				end
				addParamData(0,tr.to_a[11].to_f.to_s)

				transDE = @seqDE
				genDirectoryEntry(124,startParamLine,0,0,0,BLANK_VISIBLE,SUB_INDEPENDANT,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);

				f.set_attribute("IGES","DE",@seqDE.to_s)


				angle = m1.vector_to(p1).angle_between(m1.vector_to(p2)).radians
				m1.transform!(tr);
				p1.transform!(tr);
			    p2.transform!(tr);
				# Create param entry
				startParamLine = @seqP
				addParamData(100,m1[2].to_mm.to_f.to_s)
				addParamData(100,m1[0].to_mm.to_f.to_s)
				addParamData(100,m1[1].to_mm.to_f.to_s)
				addParamData(100,p1[0].to_mm.to_f.to_s)
				addParamData(100,p1[1].to_mm.to_f.to_s)
				addParamData(100,p2[0].to_mm.to_f.to_s)
				addParamData(0,p2[1].to_mm.to_f.to_s)
				genDirectoryEntry(100,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0,transDE);
			end
		end
	}
	@numEdges += edgeListIndex-1
  end


  def IGESexport.entity502_Vertices(e)
	verticesListIndex = 1;
	# first count the vertices and assign indices
	e.each {|f|						# for all model entities
		if (f.typename == "Edge") && (not f.deleted?) && (not f.hidden?)   # valid face ?
		v = f.vertices[0];
		if v.get_attribute("IGES","DE") == nil
			v.set_attribute("IGES","DE",@seqDE.to_s)  		# Directory Entry of Vertices 502
			v.set_attribute("IGES","Index",verticesListIndex.to_s)  		# Directory Entry of Vertices 502
			verticesListIndex += 1;
		end
		v = f.vertices[1];
		if  v.get_attribute("IGES","DE") == nil
			v.set_attribute("IGES","DE",@seqDE.to_s)  		# Directory Entry of Vertices 502
			v.set_attribute("IGES","Index",verticesListIndex.to_s)  		# Directory Entry of Vertices 502
			verticesListIndex += 1;
		end
	end
	}
	numVertices = verticesListIndex-1

	# now generate the parameter set for the entity
	verticesListIndex = 1
	startParamLine = @seqP
	addParamData(502,numVertices.to_s)
	e.each {|f|						# for all model entities
		if (f.typename == "Edge") && (not f.deleted?) && (not f.hidden?)   # valid face ?
			v = f.vertices[0];
			if v.get_attribute("IGES","Index").to_i == verticesListIndex		# add point to Param List
				p = v.position
				addParamData(502,p[0].to_mm.to_f.to_s)
				addParamData(502,p[1].to_mm.to_f.to_s)
				if verticesListIndex < numVertices
					addParamData(502,p[2].to_mm.to_f.to_s)
				else
					addParamData(0,p[2].to_mm.to_f.to_s) # letzter Eintrag
				end
				verticesListIndex += 1;
			end
			v = f.vertices[1];
			if v.get_attribute("IGES","Index").to_i == verticesListIndex		# add point to Param List
				p = v.position
				addParamData(502,p[0].to_mm.to_f.to_s)
				addParamData(502,p[1].to_mm.to_f.to_s)
				if verticesListIndex < numVertices
					addParamData(502,p[2].to_mm.to_f.to_s)
				else
					addParamData(0,p[2].to_mm.to_f.to_s) # letzter Eintrag
				end
				verticesListIndex += 1;
			end
		end
	}
	# now generate the matching directory entry
	genDirectoryEntry(502,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,1);

  end

  def IGESexport.entities504_edgeList(e)
	edgeListIndex = 1;
	startParamLine = @seqP
	# first count the vertices and assign indices
	addParamData(504,@numEdges.to_s)
	e.each {|f|						# for all model entities
		if (f.typename == "Edge") && (not f.deleted?) && (not f.hidden?)   # valid face ?
			de = f.get_attribute("IGES","DE")  		# Directory Entry of Edge 110 or arc 100
			f.set_attribute("IGES","DE504",@seqDE.to_s)  		# Directory Entry
			f.set_attribute("IGES","Index504",edgeListIndex.to_s)  		# Directory Entry
			addParamData(504,de)	# DE of edge
			v = f.vertices[0]
			de = v.get_attribute("IGES","DE")  		# Directory Entry of Vertices 502
			addParamData(504,de)	# DE of first vertex
			ind = v.get_attribute("IGES","Index")  		# List Index of Vertices 502
			addParamData(504,ind)	# Index of first vertex
			v = f.vertices[1]
			de = v.get_attribute("IGES","DE")  		# Directory Entry of Vertices 502
			addParamData(504,de)	# DE of first vertex
			ind = v.get_attribute("IGES","Index")  		# List Index of Vertices 502
			if edgeListIndex < @numEdges
				addParamData(504,ind)	# Index of first vertex
			else
				addParamData(0,ind)	# Index of first vertex, last entry
			end
			edgeListIndex += 1;
		end
	}
	# now generate the matching directory entry
	genDirectoryEntry(504,startParamLine,0,1,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);

  end

  def IGESexport.entities508_loops(e)
	faceListIndex = 1;
	# first count the vertices and assign indices
	e.each {|f|						# for all model entities
		if (f.typename == "Face") && (not f.deleted?) && (not f.hidden?)   # valid face ?
			faceListIndex += 1;
			# Create param entry
			f.loops.each{|l|		# one DE per Loop
				edges = l.edges
				startParamLine = @seqP
				l.set_attribute("IGES","DE",@seqDE.to_s)  			# Directory Entry attatched to loop
				addParamData(508,edges.size.to_s)			# number of edges in loop
				addParamData(508,0.to_s)						# edges
				numEdges = edges.size;
				num = 1;
				lastVertex = 0;
				edges.each{|edge|
					de = edge.get_attribute("IGES","DE504")  			# Directory Entry of Edge List 504
					index = edge.get_attribute("IGES","Index504")		# List Index in 504
					addParamData(508,de.to_s)
					addParamData(508,index.to_s)
					if (l.vertices[lastVertex] == edge.vertices[0])		# edge is in loop direction
						addParamData(508,1.to_s)						# Direction Flag
					else
						addParamData(508,2.to_s)						# Direction Flag
					end
					lastVertex += 1
					addParamData(508,0.to_s)						# Isparametric Flag
					if num < numEdges
					    addParamData(508,0.to_s)					# No Parameter Space Curves
					else
					    addParamData(0,0.to_s)					# No Parameter Space Curves
					end
					num += 1;
				}
				# now generate the matching directory entry
				genDirectoryEntry(508,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,1);
			}
		end
	}
	@numFaces = faceListIndex-1

  end

  def IGESexport.entities190_surfaces(e)
	# first count the vertices and assign indices
	e.each {|f|						# for all model entities
		if (f.typename == "Face") && (not f.deleted?) && (not f.hidden?)   # valid face ?
			if f.get_attribute("IGES","type") == "cone"   		# face is a cylinder surface
				# first generate a point entry 116
				startParamLine = @seqP
                m1 = Geom::Point3d.new(f.get_attribute("IGES","m1x").to_f,f.get_attribute("IGES","m1y").to_f,f.get_attribute("IGES","m1z").to_f)
                m2 = Geom::Point3d.new(f.get_attribute("IGES","m2x").to_f,f.get_attribute("IGES","m2y").to_f,f.get_attribute("IGES","m2z").to_f)
                r1 = f.get_attribute("IGES","r1")
                r2 = f.get_attribute("IGES","r2")
				v = m1.vector_to(m2)						#  vector
				if r2 < r1 then
					v.reverse!
				end
				addParamData(116,m1[0].to_mm.to_f.to_s)
				addParamData(116,m1[1].to_mm.to_f.to_s)
				addParamData(0,m1[2].to_mm.to_f.to_s)
				pointDE = @seqDE
				genDirectoryEntry(116,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);
				# now generate a vector entry 123
				startParamLine = @seqP
				addParamData(123,v[0].to_mm.to_f.to_s)
				addParamData(123,v[1].to_mm.to_f.to_s)
				addParamData(0,v[2].to_mm.to_f.to_s)
				vecDE = @seqDE
				genDirectoryEntry(123,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);

				# now generate the cylinder entity
				startParamLine = @seqP
				if r1 == r2				# this is a cylinder
					addParamData(192,pointDE.to_s)
					addParamData(192,vecDE.to_s)
					addParamData(0,f.get_attribute("IGES","r1").to_mm.to_f.to_s)
					planeDE = @seqDE
					genDirectoryEntry(192,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);
				else			# this is a cone
					addParamData(194,pointDE.to_s)
					addParamData(194,vecDE.to_s)
					addParamData(194,r1.to_mm.to_f.to_s)
					h = v.length
					w = (r2-r1).abs
					angle = Math.atan(w/h).radians
	#				puts h,w,angle,r1,r2, "-"
					addParamData(0,angle.to_f.to_s)
					planeDE = @seqDE
					genDirectoryEntry(194,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);
				end
				f.set_attribute("IGES","PlaneDE",planeDE.to_s)  					# Directory Entry of Plane 190

			else
				# first generate a point entry 116
				startParamLine = @seqP
				p = f.vertices[0].position			# point on plane, any vertex will do
				v = f.normal						# normal vector
				addParamData(116,p[0].to_mm.to_f.to_s)
				addParamData(116,p[1].to_mm.to_f.to_s)
				addParamData(0,p[2].to_mm.to_f.to_s)
				pointDE = @seqDE
				genDirectoryEntry(116,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);
				# now generate a vector entry 123
				startParamLine = @seqP
				addParamData(123,v[0].to_mm.to_f.to_s)
				addParamData(123,v[1].to_mm.to_f.to_s)
				addParamData(0,v[2].to_mm.to_f.to_s)
				normalDE = @seqDE
				genDirectoryEntry(123,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);

				# now generate the plane entity
				startParamLine = @seqP
				addParamData(190,pointDE.to_s)
				addParamData(0,normalDE.to_s)
				planeDE = @seqDE
				genDirectoryEntry(190,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);
				f.set_attribute("IGES","PlaneDE",planeDE.to_s)  					# Directory Entry of Plane 190
			end
		end
	}
  end


  def IGESexport.entities510_faces(e)
	# first count the vertices and assign indices
	e.each {|f|						# for all model entities
		if (f.typename == "Face") && (not f.deleted?) && (not f.hidden?)   # valid face ?

			surfDE = f.get_attribute("IGES","PlaneDE")  		# Directory Entry of Vertices 502
			startParamLine = @seqP
			addParamData(510,surfDE.to_s)					# SURF
			numLoops = f.loops.size
			addParamData(510,numLoops.to_s)					# number of loops
#			addParamData(510,1.to_s)						# first loop is outer loop
			addParamData(510,0.to_s)						# outer loop unspecified
			num = 1
			f.loops.each{|l|
			    loopDE = l.get_attribute("IGES","DE")  		# Directory Entry of Loop 508
				if num == numLoops
					addParamData(0,loopDE.to_s)							# last loop
				else
					addParamData(510,loopDE.to_s)						# loop entry
				end
				num += 1

			}
			f.set_attribute("IGES","FaceDE",@seqDE.to_s)  					# Directory Entry of Face 510
			genDirectoryEntry(510,startParamLine,0,0,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);
		end
	}
  end

  def IGESexport.entities514_FaceList(e)
	faceListIndex = 1;
	startParamLine = @seqP
	# first count the vertices and assign indices
	addParamData(514,@numFaces.to_s)
	e.each {|f|						# for all model entities
		if (f.typename == "Face") && (not f.deleted?) && (not f.hidden?)   # valid face ?
			de = f.get_attribute("IGES","FaceDE")  		# Directory Entry of Face 510
			addParamData(514,de)		# DE of face
			if faceListIndex < @numFaces
				addParamData(514,1.to_s)	# Orientation flag
			else
				addParamData(0,1.to_s)		# Orientation flag, last entry
			end
			faceListIndex += 1;
		end
	}
	# now generate the matching directory entry
	@faceListDE = @seqDE;
	genDirectoryEntry(514,startParamLine,0,1,0,BLANK_VISIBLE,SUB_PHYS_DEPEND,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);
  end


  def IGESexport.genManifold_186
	startParamLine = @seqP
	addParamData(186,@faceListDE.to_s)
	addParamData(186,1.to_s)
	addParamData(0,0.to_s)
	genDirectoryEntry(186,startParamLine,0,0,0,BLANK_VISIBLE,SUB_INDEPENDANT,USE_GEOMETRY,HIRARCHY_TOPDOWN,@seqP-startParamLine,0);
  end



end


   add_separator_to_menu("Tools")
   UI.menu("Tools").add_item("IGES Export") { IGESexport.export2IGES() }


# add methods to the sketchup face class
class Sketchup::Face

	# test this face if it is part of a sphere or cone
	def  checkGeometryCone
		return if get_attribute("IGES","type") == "sphere"
		smoothedges = Array.new									# only consider smoothed faces for interpolationC:\Programme\Google\Google SketchUp 7\PlugIns
		numedges = 0											# the number of smoothed edges determines the possible type
		edges.each{|e|
			numedges +=1 if e.smooth?							# count the number of smoothed edges of the face
		}
		if numedges == 1									# There is at least one smoothed edge so this is a curved shape.Use spline interpolation
		   set_attribute("IGES","type","curved")				# set the type into the attributes, at least it can be interpolated as free form
		end
		if (numedges == 2) or (numedges == 1)					# if there are two smoothed edges, the face might be part of a cone or cylinder
		   set_attribute("IGES","type","curved")				# set the type into the attributes, at least it can be interpolated as free form
		   if edges.size == 4  								# if there are two smoothed and 4 overall edges it might be a cylinder or a truncated cone
															# this type is characterized by a circular shape parallel to the smoothed line.
															# in this case, two non-smoothed edges emanate from the respective end points
															# this two edges then are secants to a circle and the midpoints of the circles
															# are connected by a normal to both circles
			   smoothedges.clear
			   edges.each{|e|								# collect the smoothed edges into a container
					smoothedges.insert(-1,e) if e.smooth?
				}
			    if numedges > 1									# There is at least one smoothed edge so this is a curved shape.Use spline interpolation
					if (smoothedges[0].vertices.include?(smoothedges[1].vertices[0])) or (smoothedges[0].vertices.include?(smoothedges[1].vertices[1]))
						return 									# smoothed edges have a vertex in common - no cylinder or cone
					end
				end
				e = smoothedges[0]							# first smoothed edge
				fs = e.faces 								# this are the two faces that are connected through this edge
				return nil if fs.size < 2 			  		# if less than two faces this is not cylindrical
				v1 = e.vertices[0]							# first end point of first edge
				e11 = nil;
				e21 = nil;									# search for the two secants there
				fs[0].vertices.each{|v|
					ev  = v.common_edge(v1)
					if (ev != nil) and (ev != e) and (v != v1)
						e11 = ev							# this is the first secant
					end
				}
				fs[1].vertices.each{|v|
					ev  = v.common_edge(v1)
					if (ev != nil) and (ev != e) and (v != v1)
						e21 = ev							# this is the second secant
					end
				}
				plane1 = e11.find_center_plane() if e11				# plane in the middle of of e11
				plane2 = e21.find_center_plane() if e21				# plane in the middle of of e21
				midline = Geom.intersect_plane_plane(plane1, plane2)	# the intersection of the two planes is a line through the midpoint

				return nil if not midline			   		# no mid-line found this is not a cylindrical shape

				m1 = v1.position.project_to_line(midline)	# mid point of end circle 1
				r1 = m1.vector_to(v1).length				# radius of end circle 1

				v2 = e.vertices[1]							# now we calculate the other end circle of the edge
				e31 = nil;
				e41 = nil;
				fs[0].vertices.each{|v| 					# search for the two secants on side 2
					ev  = v.common_edge(v2)
					if (ev != nil) and (ev != e) and (v != v2)
						e31 = ev							# this is the first secant
					end
				}
				fs[1].vertices.each{|v|
					ev  = v.common_edge(v2)
					if (ev != nil) and (ev != e) and (v != v2)
						e41 = ev							# this is the second secant
					end
				}
				plane3 = e31.find_center_plane() if e31				# plane in the middle of of e31
				plane4 = e41.find_center_plane() if e41				# plane in the middle of of e41
				midline2 = Geom.intersect_plane_plane(plane3, plane4)	# the intersection of the two planes is a line through the midpoint
				return nil if not midline2					# no mid-line found this is not a cylindrical shape
				m2 = v2.position.project_to_line(midline2) 	# mid point of end circle 2
				r2 = m2.vector_to(v2).length				# radius of end circle 2

				vm = m1.vector_to(m2)						# vector from one mid point to the other
				h = vm.length								# height of cylinder / truncated cone
				vm.normalize!								# unit vector
				ve = m1.vector_to(v1).normalize				# unit radius vector

				if (midline[1] == midline2[1])				# it is required, that the mid-lines are identical

					set_attribute("IGES","type","cone") 	# detected cone/cyl, correct the type attribute
					set_attribute("IGES","m1x",m1.x)		# add the parameters of this interpolation type
					set_attribute("IGES","m1y",m1.y)
					set_attribute("IGES","m1z",m1.z)
					set_attribute("IGES","m2x",m2.x)
					set_attribute("IGES","m2y",m2.y)
					set_attribute("IGES","m2z",m2.z)
					set_attribute("IGES","r1",r1)
					set_attribute("IGES","r2",r2)
					e11.set_attribute("IGES","type","arc") 	# add arc parameters to edge
					e11.set_attribute("IGES","mx",m1.x)
					e11.set_attribute("IGES","my",m1.y)
					e11.set_attribute("IGES","mz",m1.z)
					e11.set_attribute("IGES","r",r1)
					e31.set_attribute("IGES","type","arc") 	# add arc parameters to edge
					e31.set_attribute("IGES","mx",m2.x)
					e31.set_attribute("IGES","my",m2.y)
					e31.set_attribute("IGES","mz",m2.z)
					e31.set_attribute("IGES","r",r2)
					e21.set_attribute("IGES","type","arc") 	# add arc parameters to edge
					e21.set_attribute("IGES","mx",m1.x)
					e21.set_attribute("IGES","my",m1.y)
					e21.set_attribute("IGES","mz",m1.z)
					e21.set_attribute("IGES","r",r1)
					e41.set_attribute("IGES","type","arc") 	# add arc parameters to edge
					e41.set_attribute("IGES","mx",m2.x)
					e41.set_attribute("IGES","my",m2.y)
					e41.set_attribute("IGES","mz",m2.z)
					e41.set_attribute("IGES","r",r2)
				end
			end

			if edges.size == 3  			# if the face has three edges it might be part of a tipped cone
				set_attribute("IGES","type","curved")	# in any case it may be interpolated with method 1
				numsedges = 0
				smoothedges = []
				edges.each{|e|   							# count the smoothed edges
					if e.smooth? == true
						numsedges +=1
						smoothedges.insert(-1,e)			# add smoothed edges to a container
					end
				}

				if not (numsedges == 2) then return nil;  end  # valid cones may must have two smoothed edges
				e1 = smoothedges[0]				# first edge
				e2 = smoothedges[1] 			# second edge
				fs = nil
				# in a complete cone the two edges have one common point
				if (e1.vertices[0] == e2.vertices[0]) or (e1.vertices[0] == e2.vertices[1])	# tip is at 0 end
					m2 = e1.vertices[0].position		# tip point is a center of a circle
					r2 = 0								# with radius 0
				    fs = e1.faces 						# the faces that share edge e1
				    v1 = e1.vertices[1]					# this is the other end point
			    end
				if (e1.vertices[1] == e2.vertices[1]) or (e1.vertices[1] == e2.vertices[0])	# tip is at 1 end
					m2 = e1.vertices[1].position		# tip point is a center of a circel
					r2 = 0								# with radius 0
				    fs = e1.faces 						# the faces that share edge e1
				    v1 = e1.vertices[0]					# this is the other end point
				end

				if not fs then return nil; end			# no valid cone
				e11 = nil;								# now handle the other side, calculate the circle there as above
				e21 = nil;
				fs[0].vertices.each{|v|
					ev  = v.common_edge(v1)
					if (ev != nil) and (ev != e1) and (v != v1)
						e11 = ev						# this is the first secant
					end
				}
				fs[1].vertices.each{|v|
					ev  = v.common_edge(v1)
					if (ev != nil) and (ev != e1) and (v != v1)
						e21 = ev						# this is the second secant
					end
				}
				if (not e11) or (not e21) 				# not two secants found
					return nil							# no valid cone
				else
					plane1 = e11.find_center_plane() if e11 		# plane in the middle of secant 1
					plane2 = e21.find_center_plane() if e21		# plane in the middle of secant 2
					midline = Geom.intersect_plane_plane(plane1, plane2)	# planes intersect in the midline
					if not midline then
						return nil;						# no mid-line, no valid cone
					end
					m1 = v1.position.project_to_line(midline) 	# center of end circle
					r1 = m1.vector_to(v1).length				# radius
					ve = m1.vector_to(v1).normalize				# unit vector
				end
				vm = m1.vector_to(m2)							# vector from center to tip
				h = vm.length									# cone height
				vm.normalize!									# unit vector
				vm1 = midline[1];								# the tip of the cone also must be on the mid line
				vm1.normalize!									# unit vector

				if not ((vm1 == vm) or (vm1 == vm.reverse)) then
					return nil;
				end							# for a valid cone both vectors must be in same line

				set_attribute("IGES","type","cone") 			# is a valid cone, correct the type
				set_attribute("IGES","m1x",m1.x)				# add the cone parameters
				set_attribute("IGES","m1y",m1.y)
				set_attribute("IGES","m1z",m1.z)
				set_attribute("IGES","m2x",m2.x)
				set_attribute("IGES","m2y",m2.y)
				set_attribute("IGES","m2z",m2.z)
				set_attribute("IGES","r1",r1)
				set_attribute("IGES","r2",r2)
				e11.set_attribute("IGES","type","arc") 	# add arc parameters to edge
				e11.set_attribute("IGES","mx",m1.x)
				e11.set_attribute("IGES","my",m1.y)
				e11.set_attribute("IGES","mz",m1.z)
				e11.set_attribute("IGES","r",r1)
				e21.set_attribute("IGES","type","arc") 	# add arc parameters to edge
				e21.set_attribute("IGES","mx",m1.x)
				e21.set_attribute("IGES","my",m1.y)
				e21.set_attribute("IGES","mz",m1.z)
				e21.set_attribute("IGES","r",r1)
			end # end tipped cone
		end

	end
end


# add methods to the sketchup edge class
class Sketchup::Edge
	def find_center_plane
	  # Ebene durch 2. Kantenmittelpunkt
		   p1 = vertices[0].position
		   p2 = vertices[1].position
		   v12 = p1.vector_to(p2)
		   v12.length= v12.length/2
		   pm = p1.offset(v12)
		   plane = [pm,v12]
		#   print "Ebene: ", plane.inspect, " ",v12,"\n"
		return plane
	end
end

